//! @file IrisClassDatasets.cc
//! @brief 機械学習試験用アヤメデータセットクラス(分類確率ベクトル版)
//!
//! 機械学習の試験用ための，アヤメの「ガク，花びらの計測値」と「アヤメの種類の分類」のデータセットを集めたクラス
//! Softmax関数用の分類確率ベクトル版
//!
//! @date 2020/03/05
//! @author Yuki YOKOKURA
//
// Copyright (C) 2011-2020 Yuki YOKOKURA
// This program is free software;
// you can redistribute it and/or modify it under the terms of the FreeBSD License.
// For details, see the License.txt file.

#ifndef IRISCLASSDATASETS
#define IRISCLASSDATASETS

#include <cassert>
#include <array>
#include "Matrix.hh"
#include "BatchProcessor.hh"
#include "Shuffle.hh"

// ARCS組込み用マクロ
#ifdef ARCS_IN
	// ARCSに組み込まれる場合
	#include "ARCSassert.hh"
	#include "ARCSeventlog.hh"
#else
	// ARCSに組み込まれない場合
	#define arcs_assert(a) (assert(a))
	#define PassedLog()
	#define EventLog(a)
	#define EventLogVar(a)
#endif

namespace ARCS {	// ARCS名前空間
	//! @brief 機械学習試験用データセットクラス
	class IrisClassDatasets {
		public:
			// データセット情報の定義
			static constexpr size_t N = 4;	//!< 計測データチャネル数
			static constexpr size_t D = 150;//!< データセット数
			static constexpr size_t C = 3;	//!< 分類データチャネル数

			// アヤメの分類の定義
			Matrix<N,D> IrisMeas;	//!< 前処理済みのアヤメの計測データ
			Matrix<C,D> IrisClass;	//!< 前処理済みのアヤメの分類データ

			//! @brief コンストラクタ
			IrisClassDatasets()
				: IrisMeas(), IrisClass(), Shfl()
			{
				IrisMeas.LoadArray(IrisMeasurement);		// 計測データを行列として読み込み
				IrisClass.LoadArray(IrisClassification);	// 分類データを行列として読み込み
				ShuffleDataSets();							// シャッフルの実行
			}

			//! @brief ムーブコンストラクタ
			//! @param[in]	r	右辺値
			IrisClassDatasets(IrisClassDatasets&& r)
				: IrisMeas(r.IrisMeas), IrisClass(r.IrisClass), Shfl()
			{
				
			}
			
			//! @brief デストラクタ
			~IrisClassDatasets(){
				
			}
			
			//! @brief 計測データを表示する関数
			void DispMeasurementData(void){
				PrintMatrix(IrisMeas, "%3.1f");
			}
			
			//! @brief 分類データを表示する関数
			void DispClassificationData(void){
				PrintMatrix(IrisClass, "%1.0f");
			}

			//! @brief i番目のミニバッチ計測データを取得する関数
			//! @tparam	M	ミニバッチ数
			//! @param[out]	Y	計測データ
			//! @param[in]	i	ミニバッチ番号
			template <size_t M>
			void GetMeasurementData(Matrix<M,N>& Y, const size_t i){
				size_t k = 1;
				for(size_t j = (i - 1)*M; j < (i - 1)*M + M; j++){
					setcolumn(Y, tp(getrow(IrisMeas, j + 1)), k);	// ミニバッチサイズごとに切り出す
					++k;
				}
			}
			
			//! @brief i番目のミニバッチ分類データを取得する関数
			//! @tparam	M	ミニバッチ数
			//! @param[out]	y	訓練データ
			//! @param[in]	i	ミニバッチ番号
			template <size_t M>
			void GetClassificationData(Matrix<M,C>& y, const size_t i){
				size_t k = 1;
				for(size_t j = (i - 1)*M; j < (i - 1)*M + M; j++){
					setcolumn(y, tp(getrow(IrisClass, j + 1)), k);	// ミニバッチサイズごとに切り出す
					++k;
				}
			}
			
			//! @brief データセットをシャッフルする関数
			void ShuffleDataSets(void){
				Shfl.ShuffleMatrixRow(IrisMeas, IrisClass);	// シャッフルの実行
			}
			
		private:
			IrisClassDatasets(const IrisClassDatasets&) = delete;					//!< コピーコンストラクタ使用禁止
			const IrisClassDatasets& operator=(const IrisClassDatasets&) = delete;//!< 代入演算子使用禁止
			Shuffle Shfl;	//!< メルセンヌ・ツイスタ・シャッフル
			
			//! @brief アヤメの生計測データ
			//! {ガクの長さ [cm], ガクの幅 [cm], 花びらの長さ [cm], 花びらの幅 [cm]}
			static const std::array<std::array<double, N>, D> IrisMeasurement;
			
			//! @brief アヤメの生分類データ
			static const std::array<std::array<double, C>, D> IrisClassification;
	};
}

#endif

